from gradescopecalendar.gradescope.pyscope import GSConnection
from gradescopecalendar.calendars.ical import ICal
from gradescopecalendar.calendars.gcal import GCal
import logging


logger = logging.getLogger(__name__)


class GradescopeCalendar:
    """Interface for interacting with Gradescope and calendar applications.

    Attributes
    ----------
    username : str
        email address to login as
    password : str
        password of the account
    DEBUG : bool
        controls whether additional debug info is printed
    assignments_all : dict[]
        collection of all assignments from all courses on Gradescope

    Methods
    -------
    write_to_ical()
        creates an iCalendar file (.ics) of all assignment details
    write_to_gcal()
        connects to Google Calendar API and updates or creates Gradescope assignments
    """

    def __init__(self, email: str, password: str) -> None:
        self.assignments_all = {}
        self._get_calendar_info(email, password)

    def _get_calendar_info(self, email: str, password: str) -> None:
        """Connect to Gradescope and get assignment information."""

        # TODO: Cache assignment details in file to reduce requests to Gradescope?
        #       Might not be necessary since course page lists all assignment details
        #       so only 1 request is made per course

        # Login to Gradescope
        session = GSConnection(email, password)

        session.account.add_courses_in_account()

        # Dictionary of all assignments current in the calendar
        self.assignments_all = {}
        # Loop through all course and assignments and make Google Calendar events
        for cnum in session.account.courses:
            course = session.account.courses[cnum]
            course._load_assignments()

            # Loop through all the assignments and save them
            for assignment in course.assignments.values():
                name = f"{assignment.name} - {assignment.course.name}"
                self.assignments_all[name] = assignment

            logger.debug(f"Done parsing course on Gradescope for: {course.name}")

    def write_to_ical(self, path: str = None) -> str:
        ical = ICal()
        ical.write_to_ical(self.assignments_all, path)

    def write_to_gcal(self) -> None:
        gcal = GCal()
        gcal.write_to_gcal(self.assignments_all)



import requests
import re
from bs4 import BeautifulSoup
from gradescopecalendar.gradescope.assignment import GSAssignment


class GSCourse:
    """A class used to govern Gradescope courses.

    Attributes
    ----------
    name : str
        the full name of the course
    short_name : str
        the short name of the course
    cid : str
        6-digit course id of the course
    year : str
        the year of the course
    session : requests.Session
        the requests library Session object to manage authentication
    assignments : dict
        the available assignments in the course
    """

    def __init__(
        self, name: str, short_name: str, cid: str, year: str, session: requests.Session
    ) -> None:
        """Create a course object that has lazy eval'd assignments"""
        self.name = name
        self.course_code = short_name
        self.cid = cid
        self.year = year
        self.session = session
        self.assignments = {}
        # self._load_assignments()

    def __str__(self) -> None:
        return f"[#Course# {self.short_name} ({self.cid}) {self.name} ]"

    def _load_assignments(self) -> None:
        """Load the assignments available from the course."""

        EPOCHTIME = "1970-01-01 00:00:00 +0000"
        INVALID_ASSIGNMENT_ID = "0000000"

        assignment_resp = self.session.get(
            f"https://www.gradescope.com/courses/{self.cid}/"
        )
        parsed_assignment_resp = BeautifulSoup(assignment_resp.text, "html.parser")

        assignment_table = []
        for assignment_row in parsed_assignment_resp.findAll("tr", role="row")[
            1:
        ]:  # Skip header row
            row = []
            for th in assignment_row.findAll("th"):
                row.append(th)
            for td in assignment_row.findAll("td"):
                row.append(td)
            assignment_table.append(row)

        for row in assignment_table:
            name = row[0].text
            try:  # Assignment ID not guaranteed to be available
                aid = re.search(
                    r"/.*/assignments/(.+?)/", row[0].find("a").get("href")
                ).group(1)
            except (IndexError, AttributeError):
                aid = INVALID_ASSIGNMENT_ID
            try:  # Points not guaranteed
                points = row[1].text.split(" / ")
                points_earned = float(points[0])
                points_total = float(points[1])
                status = "Submitted"
            except (IndexError, ValueError):
                points_earned = -1
                points_total = -1
                status = row[1].text
            try:  # Open and close date not guaranteed to be available
                open_date = row[3].text if row[3].text != "" else EPOCHTIME
                close_date = row[4].text if row[4].text != "" else EPOCHTIME
            except IndexError:
                open_date = EPOCHTIME
                close_date = EPOCHTIME
            # TODO: Determine location of regrade flag
            regrades_on = False

            self.assignments[name] = GSAssignment(
                name=name,
                aid=aid,
                course=self,
                status=status,
                open_date=open_date,
                close_date=close_date,
                points=(points_earned, points_total),
                regrades_on=regrades_on,
                url=f"https://www.gradescope.com/courses/{self.cid}/assignments/{aid}/" if aid != INVALID_ASSIGNMENT_ID else "",
            )



import requests
from bs4 import BeautifulSoup
from gradescopecalendar.gradescope.account import GSAccount


class GSConnection:
    """The main connection class that keeps state about the current connection.

    Attributes
    ----------
    session : requests.Session
        the requests library Session object to manage authentication
    state : ConnState
        the state of the connection: INIT or LOGGED_IN
    account : GSAccount
        the account object created after logging into Gradescope
    """

    def __init__(self, email: str, password: str):
        """Initialize the session for the connection to Gradescope.
        
        Parameters
        ----------
        email : str
            the email address of the Gradescope account to login as
        pwd : str
            the password for the account
        """

        self.session = requests.Session()
        self.account = None
        self._login(email, password)

    def _login(self, email: str, pwd: str) -> bool:
        """Login to Gradescope using passed in credentials.

        Note: some methods depend on account privillages

        Parameters
        ----------
        email : str
            the email address of the Gradescope account to login as
        pwd : str
            the password for the account
        
        Exceptions
        ----------
        ValueError
            Invalid credentials for the Gradescope account.
        """

        # Get auth_token
        init_resp = self.session.get("https://www.gradescope.com/")
        init_resp_parsed = BeautifulSoup(init_resp.text, "html.parser")
        # TODO: simplify this loop
        for form in init_resp_parsed.find_all("form"):
            if form.get("action") == "/login":
                for inp in form.find_all("input"):
                    if inp.get("name") == "authenticity_token":
                        auth_token = inp.get("value")

        # Login to Gradescope
        login_data = {
            "commit": "Log In",
            "utf8": "✓",
            "session[email]": email,
            "session[password]": pwd,
            "session[remember_me]": 0,
            "session[remember_me_sso]": 0,
            "authenticity_token": auth_token,
        }
        login_resp = self.session.post(
            "https://www.gradescope.com/login", params=login_data
        )

        # Verify login status
        if (
            len(login_resp.history) != 0
            and login_resp.history[0].status_code == requests.codes.found
        ):
            self.account = GSAccount(self.session)
            return True
        raise ValueError("Invalid credentials.")



import logging
from pathlib import Path
from icalendar import Calendar, Event
from icalendar import vText

logger = logging.getLogger(__name__)


class ICal:
    def write_to_ical(self, assignments_all: dict, path: str = None) -> None:
        """Write assignment details to .ics file.

        Parameters
        ----------
        path : str
            the path of the output file, defaults to the script location
        """

        if not path:
            path = Path.cwd()

        # Extract relevant details from assignment for calendar event
        cal = Calendar()
        for name, assignment in assignments_all.items():
            end_time = assignment.close_date
            start_time = end_time  # Zero duration event for deadlines

            event = Event()
            event.add("summary", name)
            event.add("dtstart", start_time)
            event.add("dtend", end_time)
            event["location"] = vText(assignment.url)

            cal.add_component(event)

        # Write to file
        with open(f"{path}/gradescopecal.ics", "wb") as f:
            f.write(cal.to_ical())
            logger.info(f"Wrote file to: {path}/gradescopecal.ics")



import logging

logger = logging.getLogger(__name__)
handler = logging.StreamHandler()
handler = logging.FileHandler("gradescopecalendar.log")
formatter = logging.Formatter(
    "%(asctime)s %(name)-12s %(levelname)-8s %(message)s",
    datefmt="%a, %d %b %Y %H:%M:%S",
)
handler.setFormatter(formatter)
logger.addHandler(handler)
logger.addHandler(logging.NullHandler())



import datetime


class GSAssignment:
    """
    A class used to govern Gradescope assignments.

    Attributes
    ----------
    name : str
        the name of the assignment
    aid : str
        7-digit assignment id of the assignment
    course : GSCourse
        course object assignment is attached to
    status : str
        status of the assignment: submitted, submitted-late, no-submission, open, open-late
    open_date : str
        open date of the assignment
    close_date : str
        close date of the assignment
    points : tuple (float, float)
        tuple of points earned and total points
    regrades_on : bool
        bool of whether assignment is currently accepting regrade requests
    questions : list [str]
        list of questions in the assignment
    """

    def __init__(
        self,
        name: str,
        aid: str,
        course,
        status: str,
        open_date: str,
        close_date: str,
        url: str = None,
        points: tuple[float, float] = (0, 0),
        regrades_on: bool = False,
        questions: list[str] = None,
    ):
        """Create a GSAssignment object"""
        self.name = name
        self.aid = aid
        self.course = course
        self.status = status
        self.open_date = datetime.datetime.strptime(open_date, "%Y-%m-%d %H:%M:%S %z")
        self.close_date = datetime.datetime.strptime(close_date, "%Y-%m-%d %H:%M:%S %z")
        self.current_date = datetime.datetime.now().astimezone()
        self.url = url
        self.time_left = (
            datetime.timedelta(0)
            if self.close_date < self.current_date
            else self.current_date - self.close_date
        )
        self.points = points
        self.regrades_on = regrades_on
        self.questions = questions

    def __str__(self):
        return f"[#Assignment# {self.name} ({self.aid}) Course: {self.course.name} \t| Points: {self.points} \t {self.status}]"



from bs4 import BeautifulSoup
from gradescopecalendar.gradescope.course import GSCourse
import requests


class GSAccount:
    """A class used to govern Gradescope accounts.

    Attributes
    ----------
    session : requests.Session
        the requests library Session object to manage authentication
    courses : dict(str : GSCourse)
        dictionary using course ID as key and GSCourse as value

    Methods
    -------
    add_courses_in_account()
        adds all courses available in user account
    add_course()
    """

    def __init__(self, session: requests.Session):
        self.session = session
        self.courses = {}

    def add_courses_in_account(self) -> None:
        """Finds all courses in the current user account and adds them"""

        # Get account page and parse it using bs4
        account_resp = self.session.get("https://www.gradescope.com/account")
        account_resp_parsed = BeautifulSoup(account_resp.text, "html.parser")

        # Parameters
        ACCOUNT_COURSES_CLASS = "pageHeading"
        ACCOUNT_COURSES_HEADING = "Your Courses"
        COURSE_CLASS = "courseBox"
        COURSE_SHORTNAME_CLASS = "courseBox--shortname"
        COURSE_NAME_CLASS = "courseBox--name"

        courses = account_resp_parsed.find(
            "h1",
            class_=ACCOUNT_COURSES_CLASS,
            string=ACCOUNT_COURSES_HEADING,
        ).next_sibling

        for course in courses.find_all("a", class_=COURSE_CLASS):
            short_name = course.find("h3", class_=COURSE_SHORTNAME_CLASS).text
            name = course.find("h4", class_=COURSE_NAME_CLASS).text
            cid = course.get("href").split("/")[-1]

            year = None
            for tag in course.parent.previous_siblings:
                if tag.get("class") == "courseList--term pageSubheading":
                    year = tag.body
                    break
            self.add_course(cid=cid, name=name, short_name=short_name, year=year)

    def add_course(self, cid: str, name: str, short_name: str, year: str) -> None:
        """Creates a GSCourse object and adds it to the courses dictionary.

        Parameters
        ----------
        cid : str
            8-digit course ID
        name : str
            full name of the course
        shortname : str
            shortname of the course
        year : str
            year of the course
        """

        self.courses[cid] = GSCourse(
            cid=cid, name=name, short_name=short_name, year=year, session=self.session
        )



import time
import datetime
import os.path
import logging

from googleapiclient.discovery import build
from google_auth_oauthlib.flow import InstalledAppFlow
from google.auth.transport.requests import Request
from google.oauth2.credentials import Credentials


# If modifying these scopes, delete the file token.json.
# calendar                      See, edit, share, and permanently delete all the calendars you can access using Google Calendar
# calendar.events               View and edit events on all your calendars
# calendar.events.readonly      View events on all your calendars
# calendar.readonly             See and download any calendar you can access using your Google Calendar
# calendar.settings.readonly    View your Calendar settings
SCOPES = ["https://www.googleapis.com/auth/calendar"]

logger = logging.getLogger(__name__)


class GCal:
    """A class to handle connection with Google Calendar."""

    # API setup taken from Google docs quickstart
    def _gcal_api_setup(self):
        """Setup connection to Google Calendar API"""

        creds = None
        # The file token.json stores the user's access and refresh tokens, and is
        # created automatically when the authorization flow completes for the first
        # time.
        if os.path.exists("token.json"):
            creds = Credentials.from_authorized_user_file("token.json", SCOPES)
        # If there are no (valid) credentials available, let the user log in.
        if not creds or not creds.valid:
            if creds and creds.expired and creds.refresh_token:
                creds.refresh(Request())
            else:
                flow = InstalledAppFlow.from_client_secrets_file(
                    "credentials.json", SCOPES
                )
                creds = flow.run_local_server(port=0)
            # Save the credentials for the next run
            with open("token.json", "w") as token:
                token.write(creds.to_json())

        service = build("calendar", "v3", credentials=creds)
        return service

    def _find_gradescope_calendar(self, service) -> dict:
        """Finds or creates the Gradescope calendar and return its information."""

        # Search for existing calendar and create if needed
        user_calendars = service.calendarList().list().execute()
        gs_cal = None

        # Find existing Gradescope calendar by searching names
        for calendar in user_calendars["items"]:
            if calendar.get("summary", "") == "Gradescope":
                gs_cal = calendar
                break

        # Make a new Gradescope calendar
        if not gs_cal:
            logger.info(f"No existing Gradescope calendar, creating a new calendar...")
            gs_cal_new = {
                "kind": "calendar#calendarListEntry",
                "description": "Unofficial integration with Gradescope to integrate assignment deadlines to Google Calendar",
                "summary": "Gradescope",
            }
            gs_cal = service.calendars().insert(body=gs_cal_new).execute()
            logger.debug(f"Made a new calendar!\n{gs_cal}")
        logger.debug(f"Calendar already exists!\n{gs_cal}")
        return gs_cal

    def _get_gcal_current_assignments(self, service, gs_cal_id: str) -> dict:
        """Connects to the Gradescope calendar on Google Calendar and gets all event details.

        Parameters
        ----------
        service : googleapiclient.discovery.Resource
            resource object to interact with Google API
        gs_cal_id : str
            ID of the Gradescope calendar on Google Calendar
        """

        current_assignments = {}
        page_token = None

        # Loop through all events
        while True:
            assignment_list = (
                service.events()
                .list(calendarId=gs_cal_id, pageToken=page_token)
                .execute()
            )
            for event in assignment_list["items"]:
                name = event["summary"]
                current_assignments[name] = event
            page_token = assignment_list.get("nextPageToken")
            if not page_token:
                break

        return current_assignments

    def write_to_gcal(self, assignments_all: dict) -> bool:
        """Connects to Google Calendar API to add events for Gradescope assignments.

        Parameters
        ----------
        assignments_all : dict
            all assignments from Gradescope
        """

        # Connect to Google Calendar API service
        service = self._gcal_api_setup()

        # Search for existing calendar and create if needed
        gs_cal = self._find_gradescope_calendar(service)

        # Loop through all current assignments on Google Calendar and get their info
        current_assignments = self._get_gcal_current_assignments(service=service, gs_cal_id=gs_cal["id"])

        # Loop through all assignments from Gradescope and update/create events in Google Calendar as needed
        # Priority given to Gradescope for {open time, close time, location}
        # Priority given to Google Calendar for all other fields
        EPOCHTIME = "1970-01-01T00:00:00+0000"
        for name, assignment in assignments_all.items():
            # Format time to match time from API
            end_time = datetime.datetime.strftime(
                assignment.close_date, "%Y-%m-%dT%H:%M:%S%z"
            )
            start_time = end_time  # Zero duration event

            event_mode = None
            event_body = None
            # Update existing event details
            if name in current_assignments:
                # Attributes to compare between Google Calendar and Gradescope
                is_different_url = (
                    assignment.url != ""
                    and current_assignments[name].get("location", "") != assignment.url
                )
                is_different_start = assignment.open_date not in (
                    EPOCHTIME,
                    "",
                ) and current_assignments[name]["start"]["dateTime"] != time.strftime(
                    "%Y-%m-%dT%H:%M:%SZ", assignment.close_date.utctimetuple()
                )
                is_different_end = assignment.close_date not in (
                    EPOCHTIME,
                    "",
                ) and current_assignments[name]["end"]["dateTime"] != time.strftime(
                    "%Y-%m-%dT%H:%M:%SZ",
                    assignment.close_date.utctimetuple(),
                )

                # Only modify events with divergent urls, start, or end times to speed up execution
                if is_different_url or is_different_start or is_different_end:
                    # Check if assignment url exists and update if different from gcal
                    if is_different_url:
                        logger.debug("URL location will be updated")
                        current_assignments[name]["location"] = assignment.url
                    # Check if assignment open date exists and update if different from gcal
                    elif is_different_start:
                        logger.debug("Start time will be updated")
                        current_assignments[name]["start"]["dateTime"] = start_time
                    # Check if assignment close date exists and update if different from gcal
                    elif is_different_end:
                        logger.debug("End time will be updated")
                        current_assignments[name]["end"]["dateTime"] = end_time
                    event_mode = "update"
                    event_body=current_assignments[name]

            # Create new event details
            elif assignment.close_date != EPOCHTIME:
                logger.debug(f"Creating new event {name}")
                event_mode = "create"
                event_body = {
                    "summary": name,
                    "location": assignment.url,
                    "start": {
                        "dateTime": start_time,
                    },
                    "end": {
                        "dateTime": end_time,
                    },
                }

            if event_mode != None:
                try:
                    self._gcal_event_modify(
                        mode=event_mode,
                        service=service,
                        gs_cal_id=gs_cal["id"],
                        event_body=event_body,
                        event_id=current_assignments[name].get("id", None) if name in current_assignments else None,
                    )
                    time.sleep(0.25)  # Delay to help with rate limits
                except ValueError as e:
                    logger.exception(e)

            logger.debug(f"Google Calendar: done with assignment: {assignment.name}")

    # create new gcal event
    def _gcal_event_modify(
        self, service, gs_cal_id: str, event_body: dict, mode: str, event_id: str = None
    ) -> None:
        """Creates or updates a Google Calendar event.

        Parameters
        ----------
        mode : str
            "create", "update"
        service : googleapiclient.discovery.Resource
            resource object to interact with Google API
        gs_cal_id : str
            ID of the Gradescope calendar
        event_body : dict
            details about the calendar event to modify
        event_id : str (optional)
            eventID to update
        """

        if mode.lower() == "create":
            logger.debug(f"Creating a new gcal event...")
            event = (
                service.events().insert(calendarId=gs_cal_id, body=event_body).execute()
            )
            logger.info(f"Event created: {event.get('htmlLink')}")
        elif mode.lower() == "update":
            logger.debug(f"Updating a gcal event...")
            updated_event = (
                service.events()
                .update(calendarId=gs_cal_id, body=event_body, eventId=event_id)
                .execute()
            )
            logger.info(f"Event update: {updated_event.get('htmlLink')}")
        else:
            raise ValueError(f"Invalid mode {mode} for {event_body.get('name')}")
        logger.debug(f"Done modifying event")
